@@ -1,7 +1,7 @@
 #!/usr/bin/python
 #
 # Urwid reference documentation generation program
-#    Copyright (C) 2004-2006  Ian Ward
+#    Copyright (C) 2004-2007  Ian Ward
 #
 #    This library is free software; you can redistribute it and/or
 #    modify it under the terms of the GNU Lesser General Public
@@ -212,6 +212,7 @@ class WidgetInterface:
 		"""
 
 class ListWalkerInterface:
+
 	def get_focus(self):
 		"""
 		Returns (widget, position).
@@ -228,7 +229,8 @@ class ListWalkerInterface:
 		
 		Returns None.
 		
-		MUST be implemented.
+		MUST be implemented.  Should send "modified" signal (or call
+		self._modified if inheriting from ListWalker)
 		"""
 	
 	def get_next(self, position):
@@ -302,19 +304,26 @@ def main():
 		
 		(None,"Abstract widgets & interfaces",None),
 		(WidgetInterface, "Widget interface definition",None),
+		(urwid.Widget, "Widget",None),
 		(urwid.BoxWidget, "BoxWidget",None),
 		(urwid.FlowWidget, "FlowWidget",None),
 		(urwid.FixedWidget, "FixedWidget",None),
 		(ListWalkerInterface, "List Walker interface definition",None),
+		(urwid.ListWalker, "ListWalker", None),
 		(None,"ListBox list walkers",None),
+		(urwid.PollingListWalker, "PollingListWalker",None),
 		(urwid.SimpleListWalker, "SimpleListWalker",None),
 		
 		(None, None, None),
 		
 		(None,"Canvas painting", None),
 		(urwid.Canvas, "Canvas", None),
+		(urwid.TextCanvas, "TextCanvas", None),
+		(urwid.CompositeCanvas, "CompositeCanvas", None),
+		(urwid.SolidCanvas, "SolidCanvas", None),
 		(urwid.CanvasCombine, "CanvasCombine", None),
 		(urwid.CanvasJoin, "CanvasJoin", None),
+		(urwid.CanvasOverlay, "CanvasOverlay", None),
 		(None,"Custom formatting rules", None),
 		(urwid.TextLayout,"TextLayout", None),
 		(urwid.StandardTextLayout,"StandardTextLayout", None),
