@@ -1,7 +1,7 @@
 #!/usr/bin/python
 #
 # Urwid __init__.py 
-#    Copyright (C) 2004-2006  Ian Ward
+#    Copyright (C) 2004-2007  Ian Ward
 #
 #    This library is free software; you can redistribute it and/or
 #    modify it under the terms of the GNU Lesser General Public
@@ -21,15 +21,24 @@
 
 __all__ = [
 	'BoxWidget','Frame','Filler','ListBox','SimpleListWalker',
-	'FlowWidget','Text','Edit','IntEdit','Button','CheckBox','RadioButton',
-	'BarGraph','ProgressBar','GraphVScale',
+	'ListWalker','Signals',
+	'WidgetWrap','AttrWrap','Padding','Divider','LineBox','SolidFill',
 	'Columns','Pile','GridFlow','BoxAdapter','Overlay',
-	'AttrWrap','Padding','Divider',
-	'Canvas','CanvasCombine','CanvasJoin',
+	'FlowWidget','Text','Edit','IntEdit','Button','CheckBox','RadioButton',
+	'BarGraph','ProgressBar','GraphVScale','BigText',
+	'Canvas','CanvasCombine','CanvasJoin','CanvasCache','CompositeCanvas',
+        'TextCanvas', 'SolidCanvas',
+	'TextLayout','StandardTextLayout',
+	'set_encoding','get_encoding_mode','supports_unicode',
+	'Thin3x3Font','Thin4x3Font','HalfBlock5x4Font','HalfBlock6x5Font',
+	'HalfBlockHeavy6x5Font','Thin6x6Font','HalfBlock7x7Font',
+	'Font','get_all_fonts',
 	]
+__version__ = "0.9.8.1"
 
 
 from widget import *
 from listbox import *
 from graphics import *
-
+from canvas import *
+from font import *
