@@ -1,8 +1,8 @@
 #!/usr/bin/python
-# -* coding: utf-8 -*-
+# -*- coding: utf-8 -*-
 #
 # Urwid graphics widgets
-#    Copyright (C) 2004-2006  Ian Ward
+#    Copyright (C) 2004-2007  Ian Ward
 #
 #    This library is free software; you can redistribute it and/or
 #    modify it under the terms of the GNU Lesser General Public
@@ -27,8 +27,6 @@ from canvas import *
 from widget import *
 from escape import utf8decode
 
-try: True # old python?
-except: False, True = 0, 1
 
 class BigText(FixedWidget):
 	def __init__(self, markup, font):
@@ -41,6 +39,7 @@ class BigText(FixedWidget):
 	
 	def set_text(self, markup):
 		self.text, self.attrib = decompose_tagmarkup(markup)
+		self._invalidate()
 	
 	def get_text(self):
 		"""
@@ -50,6 +49,7 @@ class BigText(FixedWidget):
 	
 	def set_font(self, font):
 		self.font = font
+		self._invalidate()
 	
 	def pack(self, size=None, focus=False):
 		rows = self.font.height
@@ -64,7 +64,6 @@ class BigText(FixedWidget):
 		ai = ak = 0
 		o = []
 		rows = self.font.height
-		cols = 0
 		attrib = self.attrib+[(None,len(self.text))]
 		for ch in self.text:
 			if not ak:
@@ -76,14 +75,18 @@ class BigText(FixedWidget):
 				# ignore invalid characters
 				continue
 			c = self.font.render(ch)
-			c.fill_attr(a)
-			o.append(c)
-			cols += width
-			o.append(width)
-		
+			if a is not None:
+				c = CompositeCanvas(c)
+				c.fill_attr(a)
+			o.append((c, None, False, width))
 		if o:
-			return CanvasJoin(o[:-1])
-		return Canvas([""]*rows, maxcol=0)
+			canv = CanvasJoin(o)
+		else:
+			canv = TextCanvas([""]*rows, maxcol=0, 
+				check_width=False)
+			canv = CompositeCanvas(canv)
+		canv.set_depends([])
+		return canv
 		
 
 class LineBox(WidgetWrap):
@@ -119,10 +122,39 @@ class LineBox(WidgetWrap):
 		WidgetWrap.__init__(self, pile)
 
 
+class BarGraphMeta(WidgetMeta):
+	"""
+	Detect subclass get_data() method and dynamic change to
+	get_data() method and disable caching in these cases.
+
+	This is for backwards compatibility only, new programs
+	should use set_data() instead of overriding get_data().
+	"""
+	def __init__(cls, name, bases, d):
+		super(BarGraphMeta, cls).__init__(name, bases, d)
+
+		if "get_data" in d:
+			cls.render = nocache_widget_render(cls)
+			cls._get_data = cls.get_data
+		cls.get_data = property(
+			lambda self: self._get_data,
+			nocache_bargraph_get_data)
+
+def nocache_bargraph_get_data(self, get_data_fn):
+	"""
+	Disable caching on this bargraph because get_data_fn needs
+	to be polled to get the latest data.
+	"""
+	self.render = nocache_widget_render_instance(self)
+	self._get_data = get_data_fn
+
 class BarGraphError(Exception):
 	pass
 
 class BarGraph(BoxWidget):
+	__metaclass__ = BarGraphMeta
+	ignore_focus = True
+
 	eighths = utf8decode(" ▁▂▃▄▅▆▇")
 	hlines =  utf8decode("_⎺⎻─⎼⎽")
 	
@@ -229,8 +261,9 @@ class BarGraph(BoxWidget):
 			hlines = hlines[:] # shallow copy
 			hlines.sort()
 		self.data = bardata, top, hlines
+		self._invalidate()
 	
-	def get_data(self, (maxcol, maxrow)):
+	def _get_data(self, (maxcol, maxrow)):
 		"""
 		Return (bardata, top, hlines)
 		
@@ -256,6 +289,7 @@ class BarGraph(BoxWidget):
 		"""
 		assert width is None or width > 0
 		self.bar_width = width
+		self._invalidate()
 	
 	def calculate_bar_widths(self, (maxcol, maxrow), bardata ):
 		"""
@@ -465,9 +499,8 @@ class BarGraph(BoxWidget):
 		Render BarGraph.
 		"""
 		disp = self.calculate_display( (maxcol,maxrow) )
-		#assert 0, disp
 		
-		r = []
+		combinelist = []
 		for y_count, row in disp:
 			l = []
 			for bar_type, width in row:
@@ -488,9 +521,10 @@ class BarGraph(BoxWidget):
 				l.append( (a, t) )
 			c = Text(l).render( (maxcol,) )
 			assert c.rows() == 1, "Invalid characters in BarGraph!"
-			r += [c] * y_count
+			combinelist += [(c, None, False)] * y_count
 			
-		return CanvasCombine(r)
+		canv = CanvasCombine(combinelist)
+		return canv
 
 
 
@@ -681,25 +715,22 @@ class GraphVScale(BoxWidget):
 		"""
 		pl = scale_bar_values( self.pos, self.top, maxrow )
 
-		r = []
+		combinelist = []
 		rows = 0
-		blank = Canvas([""])
 		for p, t in zip(pl, self.txt):
 			p -= 1
 			if p >= maxrow: break
 			if p < rows: continue
+			c = t.render((maxcol,))
 			if p > rows:
 				run = p-rows
-				r += [blank]*run
-				rows += run
-			c = t.render((maxcol,))
+				c = CompositeCanvas(c)
+				c.pad_trim_top_bottom(run, 0)
 			rows += c.rows()
-			r.append(c)
-		if rows < maxrow:
-			r += [blank]* (maxrow-rows)
-		c = CanvasCombine(r)
-		if c.rows()>maxrow:
-			c.trim(0,maxrow)
+			combinelist.append((c, None, False))
+		c = CanvasCombine(combinelist)
+		if maxrow - rows:
+			c.pad_trim_top_bottom(0, maxrow - rows)
 		return c
 			
 			
@@ -735,6 +766,7 @@ class ProgressBar( FlowWidget ):
 		current -- current progress
 		"""
 		self.current = current
+		self._invalidate()
 	
 	def rows(self, (maxcol,), focus=False):
 		"""
@@ -759,22 +791,23 @@ class ProgressBar( FlowWidget ):
 		if self.satt is not None:
 			cs = int((cf - ccol) * 8)
 		if ccol < 0 or (ccol == 0 and cs == 0):
-			c.attr = [[(self.normal,maxcol)]]
+			c._attr = [[(self.normal,maxcol)]]
 		elif ccol >= maxcol:
-			c.attr = [[(self.complete,maxcol)]]
-		elif cs and c.text[0][ccol] == " ":
-			t = c.text[0]
+			c._attr = [[(self.complete,maxcol)]]
+		elif cs and c._text[0][ccol] == " ":
+			t = c._text[0]
 			cenc = self.eighths[cs].encode("utf-8")
-			c.text[0] = t[:ccol]+cenc+t[ccol+1:]
+			c._text[0] = t[:ccol]+cenc+t[ccol+1:]
 			a = []
 			if ccol > 0:
 				a.append( (self.complete, ccol) )
 			a.append((self.satt,len(cenc)))
 			if maxcol-ccol-1 > 0:
 				a.append( (self.normal, maxcol-ccol-1) )
-			c.attr = [a]
+			c._attr = [a]
+			c._cs = [[(None, len(c._text[0]))]]
 		else:
-			c.attr = [[(self.complete,ccol),
+			c._attr = [[(self.complete,ccol),
 				(self.normal,maxcol-ccol)]]
 		return c
 	
