@@ -1,7 +1,7 @@
 #!/usr/bin/python
 #
 # Urwid raw display module
-#    Copyright (C) 2004-2006  Ian Ward
+#    Copyright (C) 2004-2007  Ian Ward
 #
 #    This library is free software; you can redistribute it and/or
 #    modify it under the terms of the GNU Lesser General Public
@@ -23,8 +23,6 @@
 Direct terminal UI implementation
 """
 
-from __future__ import nested_scopes
-
 import fcntl
 import termios
 import os
@@ -38,15 +36,12 @@ import popen2
 import util
 import escape
 
-try: True # old python?
-except: False, True = 0, 1
-
 
 # replace control characters with ?'s
 _trans_table = "?"*32+"".join([chr(x) for x in range(32,256)])
 
 
-class Screen:
+class Screen(object):
 	def __init__(self):
 		self.palette = {}
 		self.register_palette_entry( None, 'default','default')
@@ -61,6 +56,7 @@ class Screen:
 		self.gpm_event_pending = False
 		self.last_bstate = 0
 		self.setup_G1 = True
+		self._started = False
 	
 	def register_palette( self, l ):
 		"""Register a list of palette entries.
@@ -131,6 +127,7 @@ class Screen:
 
 	def _sigwinch_handler(self, signum, frame):
 		self.resized = True
+		self.screen_buf = None
       
 	def signal_init(self):
 		"""
@@ -176,28 +173,64 @@ class Screen:
 		os.waitpid(self.gpm_mev.pid, 0)
 		self.gpm_mev = None
 	
-	def run_wrapper(self,fn):
-		""" Call fn and reset terminal on exit.
+	def start(self, alternate_buffer=True):
+		"""
+		Initialize the screen and input mode.
+		
+		alternate_buffer -- use alternate screen buffer
+		"""
+		assert not self._started
+		if alternate_buffer:
+			sys.stdout.write(escape.SWITCH_TO_ALTERNATE_BUFFER)
+		self._old_termios_settings = termios.tcgetattr(0)
+		self.signal_init()
+		tty.setcbreak(sys.stdin.fileno())
+		self._alternate_buffer = alternate_buffer
+		self._input_iter = self._run_input_iter()
+		self._next_timeout = self.max_wait
+		self._started = True
+
+	
+	def stop(self):
+		"""
+		Restore the screen.
+		"""
+		self.clear()
+		if not self._started:
+			return
+		self.signal_restore()
+		termios.tcsetattr(0, termios.TCSADRAIN, 
+			self._old_termios_settings)
+		move_cursor = ""
+		if self.gpm_mev:
+			self._stop_gpm_tracking()
+		if self._alternate_buffer:
+			move_cursor = escape.RESTORE_NORMAL_BUFFER
+		elif self.maxrow is not None:
+			move_cursor = escape.set_cursor_position( 
+				0, self.maxrow)
+		sys.stdout.write( escape.set_attributes( 
+			'default', 'default') 
+			+ escape.SI
+			+ escape.SHOW_CURSOR
+			+ escape.MOUSE_TRACKING_OFF
+			+ move_cursor + "\n" + escape.SHOW_CURSOR )
+		self._input_iter = None
+		self._started = False
+
+	def run_wrapper(self, fn, alternate_buffer=True):
+		"""
+		Call start to initialize screen, then call fn.  
+		When fn exits call stop to restore the screen to normal.
+
+		alternate_buffer -- use alternate screen buffer and restore
+			normal screen buffer on exit
 		"""
-		old_settings = termios.tcgetattr(0)
 		try:
-			self.signal_init()
-			tty.setcbreak(sys.stdin.fileno())
+			self.start(alternate_buffer)
 			return fn()
 		finally:
-			self.signal_restore()
-			termios.tcsetattr(0, termios.TCSADRAIN, old_settings)
-			move_cursor = ""
-			if self.gpm_mev:
-				self._stop_gpm_tracking()
-			if self.maxrow is not None:
-				move_cursor = escape.set_cursor_position( 
-					0, self.maxrow)
-			sys.stdout.write( escape.set_attributes( 
-				'default', 'default') 
-				+ escape.SI
-				+ escape.MOUSE_TRACKING_OFF
-				+ move_cursor + "\n" + escape.SHOW_CURSOR )
+			self.stop()
 			
 	def get_input(self, raw_keys=False):
 		"""Return pending input as a list.
@@ -240,18 +273,22 @@ class Screen:
 		Mouse button release: ('mouse release', 0, 18, 13),
 		                      ('ctrl mouse release', 0, 17, 23)
 		"""
+		assert self._started
 		
-		keys, raw = self._get_input( self.max_wait )
+		self._wait_for_input_ready(self._next_timeout)
+		self._next_timeout, keys, raw = self._input_iter.next()
 		
 		# Avoid pegging CPU at 100% when slowly resizing
 		if keys==['window resize'] and self.prev_input_resize:
 			while True:
-				keys, raw2 = self._get_input(self.resize_wait)
+				self._wait_for_input_ready(self.resize_wait)
+				self._next_timeout, keys, raw2 = \
+					self._input_iter.next()
 				raw += raw2
-				if not keys:
-					keys, raw2 = self._get_input( 
-						self.resize_wait)
-					raw += raw2
+				#if not keys:
+				#	keys, raw2 = self._get_input( 
+				#		self.resize_wait)
+				#	raw += raw2
 				if keys!=['window resize']:
 					break
 			if keys[-1:]!=['window resize']:
@@ -267,48 +304,87 @@ class Screen:
 		if raw_keys:
 			return keys, raw
 		return keys
-		
-	
-	def _get_input(self, timeout):	
-		key = self._getch(timeout)
-		raw = []
-		keys = []
-		
-		while key >= 0 or self.gpm_event_pending:
-			if self.gpm_event_pending:
-				keys += self._encode_gpm_event()
 
-			if key >= 0:
-				raw.append(key)
-				keys.append(key)
-			key = self._getch_nodelay()
+	def get_input_descriptors(self):
+		"""
+		Return a list of integer file descriptors that should be
+		polled in external event loops to check for user input.
 
-		processed = []
+		Use this method if you are implementing yout own event loop.
+		"""
+		fd_list = [sys.stdin.fileno()]
+		if self.gpm_mev is not None:
+			fd_list.append(self.gpm_mev.fromchild.fileno())
+		return fd_list
 		
+	def get_input_nonblocking(self):
+		"""
+		Return a (next_input_timeout, keys_pressed, raw_keycodes) 
+		tuple.
+
+		Use this method if you are implementing your own event loop.
 		
-		def more_fn(raw=raw,
-				# hide warnings in python 2.1
-				self_getch=self._getch, 
-				self_complete_wait = self.complete_wait):
-			key = self_getch(self_complete_wait)
-			if key >= 0:
-				raw.append(key)
-			return key
+		When there is input waiting on one of the descriptors returned
+		by get_input_descriptors() this method should be called to
+		read and process the input.
+
+		This method expects to be called in next_input_timeout seconds
+		(a floating point number) if there is no input waiting.
+		"""
+		assert self._started
+
+		return self._input_iter.next()
+
+	def _run_input_iter(self):
+		while True:
+			processed = []
+			codes = self._get_gpm_codes() + \
+				self._get_keyboard_codes()
+
+			original_codes = codes
+			try:
+				while codes:
+					run, codes = escape.process_keyqueue(
+						codes, True)
+					processed.extend(run)
+			except escape.MoreInputRequired:
+				k = len(original_codes) - len(codes)
+				yield (self.complete_wait, processed,
+					original_codes[:k])
+				original_codes = codes
+				processed = []
+
+				codes += self._get_keyboard_codes() + \
+					self._get_gpm_codes()
+				while codes:
+					run, codes = escape.process_keyqueue(
+						codes, False)
+					processed.extend(run)
 			
-		while keys:
-			run, keys = escape.process_keyqueue(keys, more_fn)
-			processed += run
+			if self.resized:
+				processed.append('window resize')
+				self.resized = False
 
-		if self.resized:
-			processed.append('window resize')
-			self.resized = False
-			self.screen_buf = None
+			yield (self.max_wait, processed, original_codes)
 
-		return processed, raw
-		
-	def _getch(self, timeout):
+	def _get_keyboard_codes(self):
+		codes = []
+		while True:
+			code = self._getch_nodelay()
+			if code < 0:
+				break
+			codes.append(code)
+		return codes
+
+	def _get_gpm_codes(self):
+		codes = []
+		while self.gpm_event_pending:
+			codes.extend(self._encode_gpm_event())
+		return codes
+
+	def _wait_for_input_ready(self, timeout):
 		ready = None
-		fd_list = [0]
+		fd_list = [sys.stdin.fileno()]
 		if self.gpm_mev is not None:
 			fd_list += [ self.gpm_mev.fromchild ]
 		while True:
@@ -322,11 +398,15 @@ class Screen:
 				if self.resized:
 					ready = []
 					break
+		return ready	
+		
+	def _getch(self, timeout):
+		ready = self._wait_for_input_ready(timeout)
 		if self.gpm_mev is not None:
-			if self.gpm_mev.fromchild in ready:
+			if self.gpm_mev.fromchild.fileno() in ready:
 				self.gpm_event_pending = True
-		if 0 in ready:
-			return ord(os.read(0, 1))
+		if sys.stdin.fileno() in ready:
+			return ord(os.read(sys.stdin.fileno(), 1))
 		return -1
 	
 	def _encode_gpm_event( self ):
@@ -402,6 +482,9 @@ class Screen:
 	
 	def draw_screen(self, (maxcol, maxrow), r ):
 		"""Paint screen with rendered canvas."""
+		assert self._started
+
+		assert maxrow == r.rows()
 
 		if self.setup_G1:
 			try:
@@ -426,31 +509,27 @@ class Screen:
 			osb = []
 		sb = []
 
-		lines = r.text
-		line = r.text
 		ins = None
-		cy = 0
-		for y in range(len(lines)):
-			line = lines[y].translate( _trans_table )
-			attr_cs = util.rle_product(r.attr[y], r.cs[y])
-			
-			if osb and osb[y] == (line, attr_cs):
+		#cy = 0
+		y = -1
+		for row in r.content():
+			y += 1
+			if osb and osb[y] == row:
 				sb.append( osb[y] )
 				continue
-			sb.append( (line, attr_cs) )
+			sb.append(row)
 			
-			if cy != y:  
-				o.append( escape.set_cursor_position(0,y) )
-			cy = y+1 # will be here after updating this line
+			#if cy != y:  
+			o.append( escape.set_cursor_position(0,y) )
+			#cy = y+1 # will be here after updating this line
 			
 			if y == maxrow-1:
-				line, attr_cs, ins = self._last_row(
-					maxcol, line, attr_cs)
+				row, back, ins = self._last_row(row)
 
-			col = 0
 			first = True
 			lasta = lastcs = None
-			for (a,cs), run in attr_cs:
+			for (a,cs, run) in row:
+				run = run.translate( _trans_table )
 				assert self.palette.has_key(a), `a`
 				if first or lasta != a:
 					o.append( self.palette[a][0] )
@@ -462,11 +541,10 @@ class Screen:
 					else:
 						o.append( escape.SO )
 					lastcs = cs
-				o.append( line[col:col+run] )
-				col += run
+				o.append( run )
 				first = False
 			if ins:
-				insert, (inserta, insertcs), back = ins
+				(inserta, insertcs, inserttext) = ins
 				ias = self.palette[inserta][0]
 				assert insertcs in [None, "0"], `insertcs`
 				if cs is None:
@@ -475,7 +553,7 @@ class Screen:
 					icss = escape.SO
 				o += [	"\x08"*back, 
 					ias, icss,
-					escape.INSERT_ON, insert,
+					escape.INSERT_ON, inserttext,
 					escape.INSERT_OFF ]
 
 		if r.cursor is not None:
@@ -501,30 +579,61 @@ class Screen:
 				raise
 
 		self.screen_buf = sb
+		self.keep_cache_alive_link = r
 				
 	
-	def _last_row(self, maxcol, line, attr_cs):
+	def _last_row(self, row):
 		"""On the last row we need to slide the bottom right character
 		into place. Calculate the new line, attr and an insert sequence
-		to do that."""
+		to do that.
 		
-		ll = len(line)
-		if ll == 0:
-			return line, attr, None
-		# offset of last position
-		i1, p1 = util.calc_text_pos( line, 0, ll, maxcol-1 )
-		if i1 == 0:
-			return line, attr, None
-		# offset of next-to-last position
-		i2, p2 = util.calc_text_pos( line, 0, ll, p1-1 )
-
-		insert = line[i2:i1] # we will insert this range
-		inserta_cs = util.rle_subseg(attr_cs,i2,i1)[0][0]
-
-		attr_cs = util.rle_subseg(attr_cs, 0, i2) + util.rle_subseg(
-			attr_cs, i1, ll)
-		line = line[0:i2] + line[i1:]
-		back = maxcol-p1
+		eg. last row:
+		XXXXXXXXXXXXXXXXXXXXYZ
 		
-		return line, attr_cs, (insert, inserta_cs, back)
+		Y will be drawn after Z, shifting Z into position.
+		"""
+		
+		new_row = row[:-1]
+		z_attr, z_cs, last_text = row[-1]
+		last_cols = util.calc_width(last_text, 0, len(last_text))
+		last_offs, z_col = util.calc_text_pos(last_text, 0, 
+			len(last_text), last_cols-1)
+		if last_offs == 0:
+			z_text = last_text
+			del new_row[-1]
+			# we need another segment
+			y_attr, y_cs, nlast_text = row[-2]
+			nlast_cols = util.calc_width(nlast_text, 0, 
+				len(nlast_text))
+			z_col += nlast_cols
+			nlast_offs, y_col = util.calc_text_pos(nlast_text, 0,
+				len(nlast_text), nlast_cols-1)
+			y_text = nlast_text[nlast_offs:]
+			if nlast_offs:
+				new_row.append((y_attr, y_cs, 
+					nlast_text[:nlast_offs]))
+		else:
+			z_text = last_text[last_offs:]
+			y_attr, y_cs = z_attr, z_cs
+			nlast_cols = util.calc_width(last_text, 0,
+				last_offs)
+			nlast_offs, y_col = util.calc_text_pos(last_text, 0,
+				last_offs, nlast_cols-1)
+			y_text = last_text[nlast_offs:last_offs]
+			if nlast_offs:
+				new_row.append((y_attr, y_cs,
+					last_text[:nlast_offs]))
+		
+		new_row.append((z_attr, z_cs, z_text))
+		return new_row, z_col-y_col, (y_attr, y_cs, y_text)
+
 			
+	
+	def clear(self):
+		"""
+		Force the screen to be completely repainted on the next
+		call to draw_screen().
+		"""
+		self.screen_buf = None
+		self.setup_G1 = True
+		
